/* opensoundcontrol.hpp ---------------------------------------------------------------//
 * 
 * (C) Copyright Ideum Inc 2012
 * 
 * This file is part of Windows 7 to TUIO Touch Utility
 *
 * Windows 7 to TUIO Touch Utility is free software: you can redistribute it
 * it under the terms of the GNU Affero General Public License as published
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Windows 7 to TUIO Touch Utility is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Windows 7 to TUIO Touch Utility.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: David Baird
 * ------------------------------------------------------------------------------------*/

#ifndef OPENSOUNDCONTROL_HPP_20110923_BNUTSDPLASDNNAGPADLAG
#define OPENSOUNDCONTROL_HPP_20110923_BNUTSDPLASDNNAGPADLAG

#include <boost/cstdint.hpp> // ...fortunately Boost saves the day :)
#include <string.h>
#include <string>
#include <vector>
#include <list>
#include <map>
#include <stdio.h>

namespace OpenSoundControl {

struct TUIO2Dcur {
    boost::int32_t s; // Session ID
    float x;
    float y;
    float X;
    float Y;
    float m;
};

struct ITUIOVisitor {
    virtual void begin_bundle() = 0;
    virtual void visit_tuio2dcur_active(std::list<boost::int32_t> session_id) = 0; // FIXME: type should be const reference
    virtual void visit_tuio2dcur_set(TUIO2Dcur & point) = 0;
    virtual void visit_tuio2dcur_fseq(boost::int32_t fseq) = 0;
    virtual void commit_bundle() = 0;
};

struct ActiveTUIOEvents : public ITUIOVisitor { // serving two purposes: active events data structure and a visitor to build the structure...
    std::list<TUIO2Dcur> tuio2dcur;
    boost::int32_t fseq;
    bool contains_tuio_2dcur_data;

    void begin_bundle() { fseq = 0; tuio2dcur.resize(0); contains_tuio_2dcur_data = false; }
    void commit_bundle() { }
    void visit_tuio2dcur_active(std::list<boost::int32_t> session_id)
    {
        contains_tuio_2dcur_data = true; 
    }
    void visit_tuio2dcur_set(TUIO2Dcur & point)
    {
        tuio2dcur.push_back(point);
    }
    void visit_tuio2dcur_fseq(boost::int32_t fseq)
    {
        contains_tuio_2dcur_data = true; 
        fseq = fseq;
    }
};

struct TUIOPrinter : public ITUIOVisitor {
    std::list<TUIO2Dcur> points;
    boost::int32_t fseq;
    void begin_bundle() { fseq = 0; points.resize(0); }
    void commit_bundle() { }
    void visit_tuio2dcur_active(std::list<boost::int32_t> session_id)
    {
    }
    void visit_tuio2dcur_set(TUIO2Dcur & point)
    {
        points.push_back(point);
    }
    void visit_tuio2dcur_fseq(boost::int32_t fseq)
    {
        fseq = fseq;
    }
    void print()
    {
        printf ("tuio: fseq=%d\n", fseq);
        for (std::list<TUIO2Dcur>::iterator i = points.begin(); i != points.end(); ++i) {
            printf ("tuio: 2dcur=0x%08x %f %f %f %f\n", i->s, i->x, i->y, i->X, i->Y, i->m);
        }
    }
};

struct OSCValue {
    enum { UNKNOWN, INT32, FLOAT32, STRING, BLOB, TIMETAG, ASAP } type;
    std::string string_value;
    boost::int32_t int32_value;
    float float32_value;
    std::string blob_value;
    boost::uint64_t timetag_value;
    int padded_size;
    static unsigned int align32(unsigned int x)
    {
        return ((x+3)>>2)<<2;
    }
    static OSCValue read_int32(const std::string & buffer)
    {
        OSCValue data;
        // Do a big-endian decode right here:
        data.type = OSCValue::INT32;
        data.int32_value = (((boost::uint8_t) buffer[0]) << 24) | (((boost::uint8_t) buffer[1]) << 16) | (((boost::uint8_t) buffer[2]) << 8) | (((boost::uint8_t) buffer[3]) << 0);
        data.padded_size = 4;
        return data;
    }
    static OSCValue read_float32(const std::string & buffer)
    {
        boost::int32_t x;
        OSCValue data;
        data.type = OSCValue::FLOAT32;
        // Assuming that localhost is using IEEE-754 floats:
        x = (((boost::uint8_t) buffer[0]) << 24) | (((boost::uint8_t) buffer[1]) << 16) | (((boost::uint8_t) buffer[2]) << 8) | (((boost::uint8_t) buffer[3]) << 0);
        memcpy(&data.float32_value, &x, 4);
        data.padded_size = 4;
        return data;
    }
    static OSCValue read_string(const std::string & buffer)
    {
        size_t i = buffer.find('\0');
        OSCValue data;
        if (i != std::string::npos) {
            data.type = OSCValue::STRING;
            data.string_value = buffer.substr(0, i);
            data.padded_size = align32(i+1); // don't forget: +1 for the '\0'!
        }
        else {
            data.type = OSCValue::UNKNOWN;
            data.padded_size = 0;
        }
        return data;
    }
    static OSCValue read_blob(const std::string & buffer)
    {
        int i = read_int32(buffer).int32_value;
        OSCValue data;
        data.type = OSCValue::BLOB;
        data.blob_value = buffer.substr(4,4+i);
        data.padded_size = align32(4+i);
        return data;
    }
    static OSCValue read_timetag(const std::string & buffer)
    {
        OSCValue data;
        data.type = OSCValue::TIMETAG;
        // ugly...
        data.timetag_value = (((boost::uint64_t)((((boost::uint8_t) buffer[0]) << 24) | (((boost::uint8_t) buffer[1]) << 16) | (((boost::uint8_t) buffer[2]) << 8) | (((boost::uint8_t) buffer[3]) << 0)))<<32) |
                             (((boost::uint64_t)((((boost::uint8_t) buffer[4]) << 24) | (((boost::uint8_t) buffer[5]) << 16) | (((boost::uint8_t) buffer[6]) << 8) | (((boost::uint8_t) buffer[7]) << 0)))<<0);
        data.padded_size = 8;
        if (data.timetag_value == 1ull) { data.type = OSCValue::ASAP; }
        return data;
    }

    static OSCValue int32(int32_t buffer)
    {
        OSCValue data;
        data.type = OSCValue::INT32;
        data.int32_value = buffer;
        data.padded_size = 4;
        return data;
    }
    static OSCValue float32(float buffer)
    {
        boost::int32_t x;
        OSCValue data;
        data.type = OSCValue::FLOAT32;
        data.float32_value = buffer;
        data.padded_size = 4;
        return data;
    }
    static OSCValue string(const std::string & buffer)
    {
        OSCValue data;
        data.type = OSCValue::STRING;
        data.string_value = buffer;
        data.padded_size = align32(buffer.size()+1); // don't forget: +1 for the '\0'!
        return data;
    }
    static OSCValue blob(const std::string & buffer)
    {
        OSCValue data;
        data.type = OSCValue::BLOB;
        data.blob_value = buffer;
        data.padded_size = align32(4+buffer.size());
        return data;
    }
    static OSCValue timetag(uint64_t buffer=1ull)
    {
        OSCValue data;
        data.type = OSCValue::TIMETAG;
        data.timetag_value = buffer;
        data.padded_size = 8;
        if (data.timetag_value == 1ull) { data.type = OSCValue::ASAP; }
        return data;
    }

    std::string write()
    {
        switch (type) {
            case INT32: return write_int32(int32_value);
            case FLOAT32: return write_float32(float32_value);
            case STRING: return write_string(string_value);
            case BLOB: return write_blob(blob_value);
            case TIMETAG: return write_timetag(1);
            case ASAP: return write_timetag(timetag_value);
            case UNKNOWN: return "";
            default: return "";
        }
    }

    std::string typetag()
    {
        switch (type) {
            case INT32: return "i";
            case FLOAT32: return "f";
            case STRING: return "s";
            case BLOB: return "b";
            //case TIMETAG: return "?";
            //case ASAP: return "?";
            case UNKNOWN: return "";
            default: return "";
        }
    }

    static std::string write_int32(int32_t buffer)
    {
        std::string data; data.resize(4);
        data[0] = (buffer >> 24) & 0xff;
        data[1] = (buffer >> 16) & 0xff;
        data[2] = (buffer >>  8) & 0xff;
        data[3] = (buffer >>  0) & 0xff;
        return data;
    }
    static std::string write_float32(float buffer)
    {
        boost::int32_t x;
        std::string data; data.resize(4);
        // Assuming that localhost is using IEEE-754 floats:
        memcpy(&x, &buffer, 4);
        data[0] = (x >> 24) & 0xff;
        data[1] = (x >> 16) & 0xff;
        data[2] = (x >>  8) & 0xff;
        data[3] = (x >>  0) & 0xff;
        return data;
    }
    static std::string write_string(const std::string & buffer)
    {
        std::string data;
        data = buffer;
        data.resize(align32(buffer.size()+1), 0); // don't forget: +1 for the '\0'!
        return data;
    }
    static std::string write_blob(const std::string & buffer)
    {
        std::string data;
        data = write_int32(buffer.size()) + buffer;
        data.resize(align32(4+buffer.size()), 0);
        return data;
    }
    static std::string write_timetag(uint64_t buffer)
    {
        std::string data; data.resize(8);
        data[0] = (buffer >> 56) & 0xff;
        data[1] = (buffer >> 48) & 0xff;
        data[2] = (buffer >> 40) & 0xff;
        data[3] = (buffer >> 32) & 0xff;
        data[4] = (buffer >> 24) & 0xff;
        data[5] = (buffer >> 16) & 0xff;
        data[6] = (buffer >>  8) & 0xff;
        data[7] = (buffer >>  0) & 0xff;
        return data;
    }
};

struct OSCMessage {
    std::string address;
    std::vector<OSCValue> values;
    int padded_size;
    static OSCMessage read_message(const std::string & buffer)
    {
        int i = 0;
        OSCMessage data;
        OSCValue typetags;
        OSCValue address_;
        address_ = OSCValue::read_string(buffer.substr(i)); i += address_.padded_size;
        data.address = address_.string_value;
        typetags = OSCValue::read_string(buffer.substr(i)); i += typetags.padded_size;
        typedef OSCValue(*decoder)(const std::string &);
        std::map<char, decoder> decoder_map;
        decoder_map['i'] = OSCValue::read_int32;
        decoder_map['f'] = OSCValue::read_float32;
        decoder_map['s'] = OSCValue::read_string;
        decoder_map['b'] = OSCValue::read_blob;
        //assert(buffer[0] == '/');
        //assert(typetags.string_value[0] == ',');
        // NOTE: begin()+1 to skip over the initial ',':
        for (std::string::iterator c = typetags.string_value.begin()+1; c != typetags.string_value.end(); c += 1) {
            OSCValue item;
            if (decoder_map.count(*c)) {
                item = decoder_map[*c](buffer.substr(i)); i += item.padded_size;
                data.values.push_back(item);
            }
            else {
                // TODO: fail
                item.type = OSCValue::UNKNOWN; i += 4;
                data.values.push_back(item);
            }
        }
        data.padded_size = i;
        return data;
    }
    std::string write()
    {
        std::string data;
        data += OSCValue::write_string(address);
        {
            std::string typetags;
            typetags += ",";
            for (int i = 0; i != values.size(); ++i) {
                typetags += values[i].typetag();
            }
            data += OSCValue::write_string(typetags);
        }
        for (int i = 0; i != values.size(); ++i) {
            data += values[i].write();
        }
        return data;
    }

};

struct OSCBundle {
    OSCValue timetag;
    std::vector<OSCMessage> messages;
    int padded_size;
    static OSCBundle read_bundle(const std::string & buffer)
    {
        int i = 0;
        OSCBundle data;
        //assert(buffer.startswith("#bundle\0"));
        i += 8; // to skip over "#bundle\0"
        data.timetag = OSCValue::read_timetag(buffer.substr(i)); i += data.timetag.padded_size;
        while (i < buffer.size()) {
            OSCValue blob;
            OSCMessage item;
            blob = OSCValue::read_blob(buffer.substr(i)); i += blob.padded_size;
            item = OSCMessage::read_message(blob.blob_value);
            data.messages.push_back(item);
        }
        data.padded_size = i;
        return data;
    }
    std::string write()
    {
        std::string data;
        data += "#bundle";
        data.resize(8, 0);
        data += timetag.write();
        for (int i = 0; i != messages.size(); ++i) {
            data += OSCValue::write_blob(messages[i].write());
        }
        return data;
    }

    static std::string write_tuio_2Dcur_active(const std::vector<TUIO2Dcur> & points, int32_t fseq_) {
        OSCBundle result;
        result.timetag = OSCValue::timetag();
        {
            OSCMessage active; active.address = "/tuio/2Dcur"; active.values.push_back(OSCValue::string("alive"));
            for (int i = 0; i < points.size(); ++i) {
                active.values.push_back(OSCValue::int32(points[i].s));
            }
            result.messages.push_back(active);
        }
        {
            for (int i = 0; i < points.size(); ++i) {
                OSCMessage set; set.address = "/tuio/2Dcur"; set.values.push_back(OSCValue::string("set"));
                set.values.push_back(OSCValue::int32(points[i].s));
                set.values.push_back(OSCValue::float32(points[i].x));
                set.values.push_back(OSCValue::float32(points[i].y));
                set.values.push_back(OSCValue::float32(points[i].X));
                set.values.push_back(OSCValue::float32(points[i].Y));
                set.values.push_back(OSCValue::float32(points[i].m));
                result.messages.push_back(set);
            }
        }
        {
            OSCMessage fseq; fseq.address = "/tuio/2Dcur"; fseq.values.push_back(OSCValue::string("fseq"));
            fseq.values.push_back(OSCValue::int32(fseq_));
            result.messages.push_back(fseq);
        }
        return result.write();
    }
};

struct OSCPacket {
    enum { UNKNOWN, BUNDLE, MESSAGE } type;
    OSCMessage message;
    OSCBundle bundle;
    int padded_size;

    static OSCPacket from_string(const std::string & buffer)
    {
        OSCPacket data;
        if (buffer[0] == '/') { // it's an OSC Message
            data.type = MESSAGE;
            data.message = OSCMessage::read_message(buffer);
            data.padded_size = data.message.padded_size;
        }
        else if(buffer[0] == '#') { // it's an OSC Bundle
            data.type = BUNDLE;
            data.bundle = OSCBundle::read_bundle(buffer);
            data.padded_size = data.bundle.padded_size;
        }
        else {
            data.type = UNKNOWN;
        }
        return data;
    }

    void accept(ITUIOVisitor & visitor) {
        if (type == BUNDLE) {
            visitor.begin_bundle();
            for (std::vector<OSCMessage>::iterator m = bundle.messages.begin(); m != bundle.messages.end(); ++m) {
                if (m->address == "/tuio/2Dcur" && m->values.size() >= 1 && m->values[0].string_value == "active") {
                    visitor.visit_tuio2dcur_active(std::list<boost::int32_t>()); // TODO
                }
                if (m->address == "/tuio/2Dcur" && m->values.size() >= 7 && m->values[0].string_value == "set") {
                    TUIO2Dcur x;
                    x.s = m->values[1].int32_value;
                    x.x = m->values[2].float32_value;
                    x.y = m->values[3].float32_value;
                    x.X = m->values[4].float32_value;
                    x.Y = m->values[5].float32_value;
                    x.m = m->values[6].float32_value;
                    visitor.visit_tuio2dcur_set(x);
                }
                if (m->address == "/tuio/2Dcur" && m->values.size() == 2 && m->values[0].string_value == "fseq") {
                    visitor.visit_tuio2dcur_fseq(m->values[1].int32_value);
                }
            }
            visitor.commit_bundle();
        }
    }
};

}

#endif /* OPENSOUNDCONTROL_HPP_20110923_BNUTSDPLASDNNAGPADLAG */
