/* w7ttu.hpp -------------------------------------------------------------------------//
 * 
 * (C) Copyright Ideum Inc 2012
 * 
 * This file is part of Windows 7 to TUIO Touch Utility
 *
 * Windows 7 to TUIO Touch Utility is free software: you can redistribute it
 * it under the terms of the GNU Affero General Public License as published
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Windows 7 to TUIO Touch Utility is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Windows 7 to TUIO Touch Utility.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: John Butler
 * ------------------------------------------------------------------------------------*/

#ifndef W7TTU_HPP
#define W7TTU_HPP

#define MESSAGE_TIMER 0x001
#define DEBUG_TIMER 0x002

#include <map>
#include <vector>
#include <set>

std::vector<long int> was_removed;
std::map<DWORD, long int> id_map;

struct addrinfo *result = NULL, *ptr = NULL, hints;

boost::int32_t frame;
long int id;
int cyscreen, cxscreen;
const char g_szClassName[] = "Win7 -> TUIO";
sockaddr_in udpAddr;
SOCKET UDPSock;
HANDLE hconsole;
const char* DEFAULT_HOST = "127.0.0.1";
const int DEFAULT_PORT = 3333;
const int MSG_PER = 16;

struct _touchpoint{
	long int id;
	DWORD win_id;
	float x;
	float y;
	void init(long int _id, float _x, float _y){
		id = _id;
		x = _x;
		y = _y;
	}
	std::string to_string(){
		return (boost::format("%d: [%d, %d], ") % id_map[id] % x % y).str();
	}
};

std::vector<_touchpoint> values;
std::vector<_touchpoint> added;
std::vector<_touchpoint> removed;

// Convert the internal point vector into a TUIO consumable packet 
// The string returned by this function can be directly sent to be consumed 
std::string points_to_tuio(){
	std::vector<OpenSoundControl::TUIO2Dcur> tuio_pts;
	for each(_touchpoint tp in values){
		OpenSoundControl::TUIO2Dcur point;
		point.s = id_map[tp.id];
		point.x = tp.x;
		point.y = tp.y;
		point.X = 0;
		point.Y = 0;
		point.m = 0;
		tuio_pts.push_back(point);
	}
	return OpenSoundControl::OSCBundle::write_tuio_2Dcur_active(tuio_pts, frame++);
}

std::string points_to_string(){
	std::string out = "";
	for each(_touchpoint tp in values){
		out += tp.to_string();
	}
	return out;
}

// Remove a _touchpoint from a vector v
// Returns true if successful, false otherwise
bool rm_value(long int rmid, std::vector<_touchpoint> &v){
	_touchpoint tp;
	for(UINT i = 0; i < v.size(); i++){
		tp = v.at(i);
		if(tp.id == rmid){
			v.erase(v.begin()+i);
			if(&v == &values){
				was_removed.push_back(tp.id);
				removed.push_back(tp);
			}
			return true;
		}
	}
	return false;
}

// Remove internal references to removed touchpoints
void clean_removed(){
	for each(_touchpoint tp in removed){
		id_map.erase(tp.id);
	}
}

// Add a new touchpoint to the internal resources
void add_value(_touchpoint tp){
	id_map[tp.id] = id++;
	added.push_back(tp);
	values.push_back(tp);
}

// Update a touchpoint's position
void update_tp(_touchpoint tp){
	for(UINT i = 0; i < values.size(); i++){
		if(values.at(i).id == tp.id){
			values.at(i).init(tp.id, tp.x, tp.y);
			return;
		}
	}
}

// Process native Windows touch events, this is called every time a WM_TOUCH event is triggered
LRESULT processTouch(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
	UINT cInputs = LOWORD(wParam);
	PTOUCHINPUT pInputs = new TOUCHINPUT[cInputs];
	bool handled = false;

	if(pInputs){
		if(GetTouchInputInfo((HTOUCHINPUT)lParam, cInputs, pInputs, sizeof(TOUCHINPUT))){
			handled = true;
		}
		else{
			// We are probably out of memory...
		}

		if(handled){
			_touchpoint tp;
			for(UINT i = 0; i < cInputs; i++){
				TOUCHINPUT ti = pInputs[i];

				// Convert screen coords to TUIO
				tp.init(ti.dwID, ((ti.x * 1.0f) / cxscreen) / 100.0f, ((ti.y * 1.0f)/ cyscreen) / 100.0f);

				// Touch up - need to remove this point
				if(ti.dwFlags & TOUCHEVENTF_UP){
					rm_value(tp.id, values);
				}
				// Touch down - new touchpoint
				else if(ti.dwFlags & TOUCHEVENTF_DOWN){
					add_value(tp);
				}
				// Point is moving / doing something
				else{
					update_tp(tp);
				}
			}
			
			CloseTouchInputHandle((HTOUCHINPUT)lParam);
			return 0;
		}
		else{
			// If we didn't handle it, send along to DefWindowProc
			return DefWindowProc(hWnd, WM_TOUCH, wParam, lParam);
		}
	}
	return 0; 
}

// Send a message with the current state of all touchpoints across the socket
void send_update(){
	std::string oscp = points_to_tuio();
	printf(oscp.c_str());

	int ret = sendto(UDPSock, oscp.c_str(), oscp.length(), 0, (struct sockaddr *)&udpAddr, sizeof(struct sockaddr));
	if(ret == SOCKET_ERROR){
		std::stringstream err;
		err << WSAGetLastError();
		MessageBox(NULL, err.str().c_str(), "!", NULL);
		WSACleanup();
		exit(1);
	}
}

#endif